#include <GUIConstantsEx.au3>
#include <StaticConstants.au3>
#include <WindowsConstants.au3>
#include <WinAPIvkeysConstants.au3>
#include <GDIPlus.au3>

Opt( "MustDeclareVars", 1 )
Opt( "WinWaitDelay", 10 )

Global Const $sIniFile = "inifile\Navigating.ini"

; Main windows, controls, handles
Global $hGui, $hPicChild, $idPicture, $hPicture, $idLblFocus

; Image and graphics variables
Global $hBitmap, $hContext, $hGraObj, $hBmpBuf, $hGraphics, $hDC, $hSrcDC

; Control program flow
Global $fHandleEvents = 0
	; 128: Move image with primary mouse button
	; 256: Resize GUI and image

Global $fXP = ( @OSVersion = "WIN_XP" )

; Project includes
#include "includes\Variables.au3"
#include "includes\NavigatingImage.au3"
#include "includes\DrawFunctions.au3"

ImageAndPixelEditor()


Func ImageAndPixelEditor()

	; --- Create GUI ---

	; Main GUI
	$hGui = GUICreate( "Navigating in an image", $iPicWidth+10, $iPicHeight+10, 500, 50, _
	                   BitXOR( BitOR( $WS_OVERLAPPEDWINDOW, $WS_CLIPCHILDREN ), $WS_MAXIMIZEBOX ) )
	$idLblFocus = GUICtrlCreateLabel( "", ($fXP?4:0), ($fXP?5:1), 14, 12 )
	GUICtrlSetResizing( $idLblFocus, $GUI_DOCKALL )
	GUIRegisterMsg( $WM_NCACTIVATE, "WM_NCACTIVATE" )
	GUISetState()

	; Child window for picture control
	$hPicChild = GUICreate( "", $iPicWidth, $iPicHeight, ($fXP?4:0), ($fXP?5:1), $WS_POPUP, $WS_EX_MDICHILD, $hGui )
	GUISetState( @SW_DISABLE, $hPicChild )
	GUISetState()

	; Picture control in child window
	$idPicture = GUICtrlCreatePic( "", 0, 0, $iPicWidth, $iPicHeight )
	GUICtrlSetResizing( $idPicture, $GUI_DOCKALL )
	$hPicture = GUICtrlGetHandle( $idPicture )
	GUISwitch( $hGui )

	; GUI width and height
	Local $aPos = WinGetPos( $hGui )
	$iWinWidth = $aPos[2]
	$iWinHeight = $aPos[3]
	$iMinWidth = $iWinWidth
	$iMinHeight = $iWinHeight

	; Windows messages
	GUIRegisterMsg( $WM_ACTIVATE,      "WM_ACTIVATE"      )
	GUIRegisterMsg( $WM_MOUSEWHEEL,    "WM_MOUSEWHEEL"    )
	GUIRegisterMsg( $WM_GETMINMAXINFO, "WM_GETMINMAXINFO" )
	GUIRegisterMsg( $WM_SIZING,        "WM_SIZING"        )

	; Catch key presses
	Local $hMessageHandler = DllCallbackRegister( "MessageHandler", "long", "int;wparam;lparam" )
	Local $hMessageHook = _WinAPI_SetWindowsHookEx( $WH_GETMESSAGE, DllCallbackGetPtr( $hMessageHandler ), 0, _WinAPI_GetCurrentThreadId() )

	; --- Load image ---

	; Start GDIPlus
	_GDIPlus_Startup()

	; Open on startup
	Local $sImage = @ScriptDir & "\images\Image.bmp", $hImage
	$hImage = _GDIPlus_BitmapCreateFromFile( $sImage )
	$iImageWidth = _GDIPlus_ImageGetWidth( $hImage )
	$iImageHeight = _GDIPlus_ImageGetHeight( $hImage )

	; Image
	$hBitmap = _GDIPlus_BitmapCreateFromScan0( $iImageWidth, $iImageHeight )
	$hContext = _GDIPlus_ImageGetGraphicsContext( $hBitmap )
	_GDIPlus_GraphicsDrawImageRect( $hContext, $hImage, 0, 0, $iImageWidth, $iImageHeight )
	_GDIPlus_ImageDispose( $hImage )

	; Device contexts
	$hDC = _WinAPI_GetDC( $hPicture )
	$hSrcDC = _WinAPI_CreateCompatibleDC( $hDC )

	; --- Draw graphics on top of the picture control using buffered drawing ---

	; Create pen to draw frame and grid
	$hPenFrameGrid = _GDIPlus_PenCreate( 0xFFCCCCCC, 1 )
	; Graphics object to draw the contents of the bitmap buffer on top of the picture control
	$hGraObj = _GDIPlus_GraphicsCreateFromHWND( $hPicture )
	; Bitmap buffer for buffered graphics drawing
	$hBmpBuf = _GDIPlus_BitmapCreateFromGraphics( $iPicWidth, $iPicHeight, $hGraObj )
	; Graphics context to draw graphics in the bitmap buffer
	$hGraphics = _GDIPlus_ImageGetGraphicsContext( $hBmpBuf )

	; Positions and sizes
	ImagePosSize()

	; Draw image
	ImageDraw( True )

	; Set focus to $idLblFocus
	; Required because $hPicChild is disabled
	GUICtrlSetState( $idLblFocus, $GUI_FOCUS )

	; --- GUI main loop ---

	Local $iMsg, $aInfo
	While 1

		$iMsg = GUIGetMsg()

		If $fHandleEvents Then

			Select
				Case BitAND( $fHandleEvents, 128 ) ; Move image with primary mouse button
					ImageMove( $aInfo[0], $aInfo[1] )

				Case BitAND( $fHandleEvents, 256 ) ; Resize GUI and image
					ImageResize()
			EndSelect

		Else

			; --- View mode ---

			Switch $iMsg

				Case $GUI_EVENT_MOUSEMOVE, 0
					ContinueLoop

				Case $GUI_EVENT_PRIMARYDOWN
					$aInfo = GUIGetCursorInfo( $hPicChild )
					If $aInfo[4] = $idPicture Then
						; - Click a point in picture control to center the point (zoom in/out around this point)
						; - Click and drag with the mouse to move the image in the picture control
						$fHandleEvents = BitOR( $fHandleEvents, 128 )
					ElseIf GUIGetCursorInfo( $hGui )[1] < 0 Then
						; Click in menu bar or title bar areas
						GUICtrlSetState( $idLblFocus, $GUI_FOCUS )
					EndIf

				Case $GUI_EVENT_RESTORE
					ImageDraw()

				Case $GUI_EVENT_CLOSE
					ExitLoop

			EndSwitch

		EndIf

	WEnd

	;_GDIPlus_ImageSaveToFile( $hBitmap, "images\Saved.bmp" )

	_WinAPI_UnhookWindowsHookEx( $hMessageHook )

	_GDIPlus_PenDispose( $hPenFrameGrid )

	_GDIPlus_BitmapDispose( $hBitmap )
	_GDIPlus_BitmapDispose( $hBmpBuf )
	_GDIPlus_GraphicsDispose( $hContext )
	_GDIPlus_GraphicsDispose( $hGraObj )
	_GDIPlus_GraphicsDispose( $hGraphics )
	_WinAPI_ReleaseDC( $hPicture, $hDC )
	_WinAPI_DeleteDC( $hSrcDC )
	_GDIPlus_Shutdown()

	GUIDelete( $hGui )
	Exit

EndFunc

; Prevent child windows in making GUI title bar dimmed
Func WM_NCACTIVATE( $hWnd, $iMsg, $wParam )
	If $hWnd = $hGui Then
		If Not $wParam Then Return 1
	EndIf
	Return $GUI_RUNDEFMSG
EndFunc
