#include-once

; Functions to
;     - calculate image positions and sizes
;     - resize image when GUI is resized
;     - zoom in/out and move image

Global Const $tagMSG = "hwnd hwnd;uint message;wparam wParam;lparam lParam;dword time;int X;int Y"
Global Const $STM_SETIMAGE = 0x0172

; Variables to handle position and size for source (image) and destination (picture control)
Global $iPicSizeStep = 96
Global $iPicWidth = $iPicSizeStep * 6
Global $iPicHeight = $iPicSizeStep * 6
Global $iImageWidth, $iImageHeight
Global $iWinWidth, $iWinHeight
Global $iMinWidth, $iMinHeight

Global $iXDest, $iYDest, $iWidthDest, $iHeightDest ; Destination (picture control) position and size
Global $iXSrc, $iYSrc, $iWidthSrc, $iHeightSrc     ; Source (image) position and size
Global $iXMiddleSrc = -1, $iYMiddleSrc = -1        ; Center destination around source middle point

Global $dx = 0, $dxMax
Global $dy = 0, $dyMax

Global $nZoom = 1, $iZoom = 5 ; Zoom factor and level, initial zoom factor $nZoom = 1 at zoom level $iZoom = 5
Global $aZoom = [ 1/8, 1/6, 1/4, 1/3, 1/2, 1, 2, 3, 4, 6, 8, 12, 16, 24, 32 ] ; 15 zoom levels with 15 factors

; Miscellaneous variables for navigation
Global $fMoveZoom = True, $fMoveImage = False
Global $fShiftFlag = 0, $fShift = False
Global $iDeactive = 0


; Position and size for source (image) and destination (picture control)
Func ImagePosSize()

	; --- Picture destination ---

	$iWidthDest = $iImageWidth * $nZoom
	Local $iEmptyWidth = $iPicWidth - $iWidthDest
	If $iEmptyWidth < 0 Then $iWidthDest = $iPicWidth

	$iHeightDest = $iImageHeight * $nZoom
	Local $iEmptyHeight = $iPicHeight - $iHeightDest
	If $iEmptyHeight < 0 Then $iHeightDest = $iPicHeight

	If Mod( $iEmptyWidth, 2 ) Then $iEmptyWidth -= 1
	$iXDest = $iEmptyWidth / 2
	If $iEmptyWidth < 0 Then _
		$iXDest = 0

	If Mod( $iEmptyHeight, 2 ) Then $iEmptyHeight -= 1
	$iYDest = $iEmptyHeight / 2
	If $iEmptyHeight < 0 Then _
		$iYDest = 0

	; --- Image source ---

	$iXSrc = 0
	$iYSrc = 0
	$iWidthSrc = $iImageWidth
	$iHeightSrc = $iImageHeight
	$dxMax = 0
	$dyMax = 0

	If $iEmptyWidth < 0 Then
		$iWidthSrc = $iPicWidth / $nZoom
		Local $iTooManyXpixels = $iImageWidth - $iWidthSrc
		$dxMax = $iTooManyXpixels
		If $iXMiddleSrc > -1 Then
			$iXSrc = $iXMiddleSrc - Int( $iWidthSrc / 2 )
			If $iXSrc + $iWidthSrc > $iImageWidth Then
				$iXSrc = $iImageWidth - $iWidthSrc
			ElseIf $iXSrc < 0 Then
				$iXSrc = 0
			EndIf
		Else
			If Mod( $iTooManyXpixels, 2 ) Then $iTooManyXpixels -= 1
			$iXSrc = $iTooManyXpixels / 2
		EndIf
	EndIf

	If $iEmptyHeight < 0 Then
		$iHeightSrc = $iPicHeight / $nZoom
		Local $iTooManyYpixels = $iImageHeight - $iHeightSrc
		$dyMax = $iTooManyYpixels
		If $iYMiddleSrc > -1 Then
			$iYSrc = $iYMiddleSrc - Int( $iHeightSrc / 2 )
			If $iYSrc + $iHeightSrc > $iImageHeight Then
				$iYSrc = $iImageHeight - $iHeightSrc
			ElseIf $iYSrc < 0 Then
				$iYSrc = 0
			EndIf
		Else
			If Mod( $iTooManyYpixels, 2 ) Then $iTooManyYpixels -= 1
			$iYSrc = $iTooManyYpixels / 2
		EndIf
	EndIf

EndFunc

Func ImageResize()
	While 1
		Switch GUIGetMsg()
			Case $GUI_EVENT_MOUSEMOVE
				ContinueLoop

			Case $GUI_EVENT_PRIMARYUP
				; Note that $GUI_EVENT_RESIZED is fired before $GUI_EVENT_PRIMARYUP.
				GUICtrlSetState( $idLblFocus, $GUI_FOCUS )
				$fHandleEvents -= 256
				$fResizing = 0
				ExitLoop

			Case $GUI_EVENT_RESIZED
				If $fResizing = 1 Then
					; Resize GUI window
					; (Child window and picture control were
					;  resized in the WM_SIZING function.)

					$fResizing = 2
					WinMove( $hGui, "", Default, Default, $iWinWidth, $iWinHeight )
					; WinMove will cause an resize event ($WM_GETMINMAXINFO and $WM_SIZING)
					; $fResizing = 2 prevents the $GUI_EVENT_RESIZED code from running twice

					; Update bitmap buffer and graphics
					_GDIPlus_BitmapDispose( $hBmpBuf )
					_GDIPlus_GraphicsDispose( $hGraObj )
					_GDIPlus_GraphicsDispose( $hGraphics )
					$hGraObj = _GDIPlus_GraphicsCreateFromHWND( $hPicture )
					$hBmpBuf = _GDIPlus_BitmapCreateFromGraphics( $iPicWidth, $iPicHeight, $hGraObj )
					$hGraphics = _GDIPlus_ImageGetGraphicsContext( $hBmpBuf )

					; Update grid in bitmap buffer
					$fGraphics = BitOR( $fGraphics, 5 ) ; (1 + 4)

					; Redraw image and graphics on top of image
					ImageDraw( $iXDest Or $iYDest )
				EndIf
		EndSwitch
	WEnd
EndFunc

Func WM_SIZING( $hWnd, $iMsg, $wParam, $lParam )
	If Not $fResizing Then
		$fHandleEvents += 256
		$fGraphics -= 1 ; Don't redraw graphics on top of image while resizing
		$fResizing = 1
	EndIf

	Local $tRECT = DllStructCreate( $tagRECT, $lParam ), $fResize = False, $dHeight = 0

	; Width
	Local $dw = DllStructGetData( $tRECT, "Right" ) - DllStructGetData( $tRECT, "Left" ) - $iWinWidth
	If $dw Then
		If $dw > $iPicSizeStep Then
			$iWinWidth += $iPicSizeStep
			$iPicWidth += $iPicSizeStep
			$fResize = True
		ElseIf $dw < 0 Then
			$iWinWidth -= $iPicSizeStep
			$iPicWidth -= $iPicSizeStep
			$fResize = True
		EndIf
	EndIf

	; Height
	Local $dh = DllStructGetData( $tRECT, "Bottom" ) - DllStructGetData( $tRECT, "Top" ) - $iWinHeight
	If $dh Then
		If $dh > $iPicSizeStep Then
			$iWinHeight += $iPicSizeStep
			$iPicHeight += $iPicSizeStep
			$dHeight += $iPicSizeStep
			$fResize = True
		ElseIf $dh < 0 Then
			$iWinHeight -= $iPicSizeStep
			$iPicHeight -= $iPicSizeStep
			$dHeight -= $iPicSizeStep
			$fResize = True
		EndIf
	EndIf

	If $fResize Then
		; Resize child window and picture control
		WinMove( $hPicChild, "", Default, Default, $iPicWidth, $iPicHeight )
		GUICtrlSetPos( $idPicture, Default, Default, $iPicWidth, $iPicHeight )

		; Redraw image
		ImagePosSize()
		ImageDraw( $iXDest Or $iYDest ) ; First redraw while resizing will clear graphics on top of image
	EndIf
EndFunc

Func WM_GETMINMAXINFO( $hWnd, $iMsg, $wParam, $lParam )
	Local $tMINMAXINFO = DllStructCreate( "int;int;int;int;int;int;int;int;int;int", $lParam )
	DllStructSetData( $tMINMAXINFO, 7, $iMinWidth )  ; Min GUI width
	DllStructSetData( $tMINMAXINFO, 8, $iMinHeight ) ; Min GUI height
EndFunc

Func WM_MOUSEWHEEL( $hWnd, $iMsg, $wParam, $lParam )
	If Not $fMoveZoom Or GUIGetCursorInfo( $hPicChild )[4] <> $idPicture Then Return
	Local $WHEEL_DELTA = _WinAPI_HiWord( $wParam )
	If $WHEEL_DELTA > 0 Then
		If $iZoom < 13 Then _
			ZoomInOut( +1 )
	Else
		If $iZoom > 0 Then _
			ZoomInOut( -1 )
	EndIf
EndFunc

Func ZoomInOut( $iInc )
	$iZoom += $iInc
	$nZoom = $aZoom[$iZoom]
	ImagePosSize()
	If BitAND( $fFrameGrid, 2 ) Then _
		$fGraphics = BitOR( $fGraphics, 4 )
	ImageDraw( $iXDest Or $iYDest, True )
EndFunc


Func ImageMove( $dx0, $dy0 )
	Local $aInfo, $fMouseMove = 0, $fDraw = False

	$fMoveImage = True

	While 1
		Switch GUIGetMsg()
			Case $GUI_EVENT_MOUSEMOVE
				If Not $fMouseMove Then
					$fMouseMove = 1
				Else
					$aInfo = GUIGetCursorInfo( $hPicChild )
					If $fMouseMove = 1 Then
						; Move (drag) image with mouse
						If 0 < $iXSrc + $dx And $iXSrc + $dx < $dxMax Then
							$dx = -Round( ( $aInfo[0] - $dx0 ) / $nZoom )
							If $iXSrc + $dx < 0 Then
								$dx = -$iXSrc
							ElseIf $iXSrc + $dx > $dxMax Then
								$dx = $dxMax - $iXSrc
							EndIf
							$fDraw = True
						ElseIf 0 < $iXSrc - Round( ( $aInfo[0] - $dx0 ) / $nZoom ) And $iXSrc - Round( ( $aInfo[0] - $dx0 ) / $nZoom ) < $dxMax Then
							$dx = -Round( ( $aInfo[0] - $dx0 ) / $nZoom )
							$dx = $dx > 0 ? 1 : -1
							$fDraw = True
						EndIf
						If 0 < $iYSrc + $dy And $iYSrc + $dy < $dyMax Then
							$dy = -Round( ( $aInfo[1] - $dy0 ) / $nZoom )
							If $iYSrc + $dy < 0 Then
								$dy = -$iYSrc
							ElseIf $iYSrc + $dy > $dyMax Then
								$dy = $dyMax - $iYSrc
							EndIf
							$fDraw = True
						ElseIf 0 < $iYSrc - Round( ( $aInfo[1] - $dy0 ) / $nZoom ) And $iYSrc - Round( ( $aInfo[1] - $dy0 ) / $nZoom ) < $dyMax Then
							$dy = -Round( ( $aInfo[1] - $dy0 ) / $nZoom )
							$dy = $dy > 0 ? 1 : -1
							$fDraw = True
						EndIf
						If $fDraw Then
							ImageDraw()
							$fDraw = False
						EndIf
					EndIf
				EndIf

			Case $GUI_EVENT_PRIMARYUP
				$fHandleEvents -= 128
				ExitLoop
		EndSwitch
	WEnd

	$fMoveImage = False

	If Not $fMouseMove Then
		; Click a point in picture control to center the point (zoom in/out around this point)
		$iXMiddleSrc = Int( $dx0 / $nZoom ) + $iXSrc - $iXDest / $nZoom ; Source coordinates
		$iYMiddleSrc = Int( $dy0 / $nZoom ) + $iYSrc - $iYDest / $nZoom
	Else
		If $fMouseMove = 1 Then
			; Click and drag with the mouse to move the image in the picture control
			$iXSrc += $dx
			$iYSrc += $dy
			$iXMiddleSrc = $iXSrc + Int( $iWidthSrc / 2 )
			$iYMiddleSrc = $iYSrc + Int( $iHeightSrc / 2 )
			$dx = 0
			$dy = 0
		EndIf
		ImageDraw()
	EndIf
EndFunc


Func MessageHandler( $nCode, $wParam, $lParam )

	; Handle keystrokes to move and zoom image
	If $fMoveZoom And GUIGetCursorInfo( $hPicChild )[4] = $idPicture Then

		Local Static $iKey = 0, $dx = 1, $dy = 1
		Local $tMSG = DllStructCreate( $tagMSG, $lParam )
		Local $iMsg = DllStructGetData( $tMSG, "message" )

		If $iMsg = $WM_KEYDOWN Then
			$iMsg = DllStructGetData( $tMSG, "wParam" )

			Switch $iMsg

				Case $VK_LEFT
					If Not $iKey Then
						$iKey = $iMsg
						$fMoveImage = True
					EndIf
					If $iXSrc < $dxMax Then
						$iXSrc += $dx
						If $iXSrc > $dxMax Then _
							$iXSrc = $dxMax
						$iXMiddleSrc = $iXSrc + Int( $iWidthSrc / 2 )
						ImageDraw()
					EndIf

				Case $VK_UP
					If Not $iKey Then
						$iKey = $iMsg
						$fMoveImage = True
					EndIf
					If $iYSrc < $dyMax Then
						$iYSrc += $dy
						If $iYSrc > $dyMax Then _
							$iYSrc = $dyMax
						$iYMiddleSrc = $iYSrc + Int( $iHeightSrc / 2 )
						ImageDraw()
					EndIf

				Case $VK_RIGHT
					If Not $iKey Then
						$iKey = $iMsg
						$fMoveImage = True
					EndIf
					If $iXSrc > 0 Then
						$iXSrc -= $dx
						If $iXSrc < 0 Then _
							$iXSrc = 0
						$iXMiddleSrc = $iXSrc + Int( $iWidthSrc / 2 )
						ImageDraw()
					EndIf

				Case $VK_DOWN
					If Not $iKey Then
						$iKey = $iMsg
						$fMoveImage = True
					EndIf
					If $iYSrc > 0 Then
						$iYSrc -= $dy
						If $iYSrc < 0 Then _
							$iYSrc = 0
						$iYMiddleSrc = $iYSrc + Int( $iHeightSrc / 2 )
						ImageDraw()
					EndIf

				Case $VK_PRIOR
					If $iZoom < 13 Then _
						ZoomInOut( +1 )

				Case $VK_NEXT
					If $iZoom > 0 Then _
						ZoomInOut( -1 )

				Case $VK_SHIFT
					Switch $fShiftFlag
						Case 0
							If Not $fShift Then
								$fShift = True
								$dx = $iWidthSrc
								$dy = $iHeightSrc
							EndIf
						Case Else
							If Not $fShift Then _
								$fShift = True
					EndSwitch

			EndSwitch

		ElseIf $iMsg = $WM_KEYUP Then

			Switch DllStructGetData( $tMSG, "wParam" )

				Case $VK_LEFT To $VK_DOWN
					$fMoveImage = False
					ImageDraw()
					$iKey = 0

				Case $VK_SHIFT
					Switch $fShiftFlag
						Case 0
							$fShift = False
							$dx = 1
							$dy = 1
						Case Else
							$fShift = False
							$fShiftFlag = 0
					EndSwitch

			EndSwitch

		EndIf

	EndIf

	; Activate or deactivate GUI?
	Local Static $iDeactivePrev = 0
	If $iDeactivePrev <> $iDeactive Then
		$iDeactivePrev = $iDeactive
		If $iDeactive Then
			$fGraphics -= 1
			ToolTip("")
		Else
			$fGraphics += 1
		EndIf       ; Deactivated: Delete graphics on top of image
		ImageDraw() ; Activated:   Draw graphics on top of image
	EndIf

EndFunc

Func WM_ACTIVATE( $hWnd, $iMsg, $wParam, $lParam )
	If BitAND( $wParam, 0xFFFF ) Then
		$iDeactive = 0 ; Activated
	Else
		$iDeactive = 1 ; Deactivated
	EndIf
	Return $GUI_RUNDEFMSG
EndFunc
